"""
assignment.py
weavebj1 - 2025

Contains tools surrounding assignment object, which relies on pygrader library

Notable features:
    - Create assignment using json file
    - Data output into .csv, .txt, pandas df, or simple dictionary
    - Use self.pypackage_command to get command line prompt needed to package 
      entire assignment into .exe

Part of python-grader library
"""

import os
import json
import datetime
import logging
from collections import defaultdict
import pandas as pd
from mupygrader import test_file
from mupygrader import utils
from mupygrader.json_interface import Problem

logging.basicConfig(format='%(asctime)s - %(levelname)s - %(message)s')

class Assignment:
    """
    Assignment object represents an assigned set of problems.

    Attributes:
        config_file (str): The name of the json file being used.
        course (str):      Name of the course (optional)
        name (str):        Name of the assignment (optional)
        late_mult (num):   Number multiplied by late assignments
        problems ([json_interface.problem]):    Objects representing each problem
        solutions (test_file.TestFile):         Objects representing solution files
    """

    def __init__(self, config_file):
        """
        Constructor for assignment object

        Parameters:
            config_file (str): The name of the json file being used
        """
        self.config_file = config_file
        self.solutions = []

        try:
            if utils.running_in_exe():
                config_file = os.path.join(utils.exe_resource_path(), config_file)

            with open(config_file, "r", encoding="utf-8") as json_file:
                data = json.load(json_file)

        except FileNotFoundError as e:
            e.add_note(f"Could not find config file {config_file}")
            raise e


        self.course = data["course"]
        self.name = data["assignment"]
        self.late_mult = data["late_mult"]

        self.problems = []
        for p in data["problems"]:
            self.problems.append(Problem(self, **p))

        for submdir in data["submission_folders"]:
            self.add_student_submissions(os.path.abspath(submdir))

    def init_resources(self):
        """
        Runs initializations for files, including running solutions
        """
        for p in self.problems:

            if p.runner_file:
                p.init_runner()

            p.init_solution()
            try:
                logging.info("Running solution for %s from %s", p.name, p.solution.filepath)
                p.solution.code.run()
            except Exception as e:
                e.add_note(f"Encountered a problem running the solution for {p.solution.filename}")
                raise e


    def add_student_submissions(self, folder):
        """
        Places student submissions into the correct problem based on the filenames

        Parameters:
            folder (str): Folder to look for submissions in
        """
        files_to_add = test_file.get_pyfiles(folder)

        for p in self.problems:
            for file in files_to_add:
                if p.file_whitelist in file:
                    p.add_submission(os.path.join(folder, file))


    def run_submissions(self):
        """
        Runs all solutions associated with all problems
        """
        self.init_resources()

        for p in self.problems:
            for sub in p.submissions:
                logging.info("Running %s for %s", p.name, sub.student_name)
                sub.run_and_score(p.solution)
                utils.close_plots()

    def to_dict(self):
        """
        Makes a dictionary that has personal identifier as a key, 
        and a list of their submissions as a value        
        """
        out_dict = defaultdict(list)
        for p in self.problems:
            for submission in p.submissions:
                out_dict[submission.identifier].append(submission)

        return out_dict

    @property
    def max_points(self):
        """
        Retreives the sum of points from all associated problems        
        """
        total = 0
        for p in self.problems:
            total += p.points
        return total

    def file_header(self):
        """
        Creates text for a header to be used in multiple output formats        
        """
        head_str = ""
        head_str += f"Assignment: {self.name}\n"
        head_str += f"Max Points: {self.max_points}\n"
        head_str += f"Report generated: {datetime.datetime.now()}\n"
        return head_str

    def output_dataframe(self):
        """
        Creates a pandas dataframe of the data        
        """
        asgmt_dict = self.to_dict()
        csv_rows = []
        for key,submissions in asgmt_dict.items():
            grade = 0
            total_points = 0
            all_canvas_submissions = True
            for s in submissions:
                grade += s.grade
                total_points += s.points_scored
                all_canvas_submissions &= s.is_canvas

            new_row = {}
            if all_canvas_submissions:
                new_row["name"] = key[0]
                new_row["sys_id"] = key[1]
            else:
                new_row["name"] = key

            new_row["grade"] = grade
            new_row["raw_points"] = total_points
            for s in submissions:
                new_row[s.params.name] = s.grade
                new_row[f"{s.params.name}_is_late"] = s.is_late
                new_row[f"{s.params.name}_raw_points"] = s.points_scored
                new_row[f"{s.params.name}_errors"] = s.error_comments

            csv_rows.append(new_row)

        return pd.DataFrame(csv_rows)


    def output_terminal(self):
        """
        Displays information about the output in the terminal        
        """
        asgmt_dict = self.to_dict()

        print(self.file_header())

        for submissions in asgmt_dict.values():

            points_earned = 0
            for s in submissions:
                points_earned += s.grade

            score_str = f"{submissions[0].student_name}:"
            if submissions[0].is_canvas: # Only print if the student is known
                score_str += f" {points_earned:0.2f}/{self.max_points:0.2f}"
                score_str += f" ({100*points_earned/self.max_points:2.0f}%)"
            print(score_str)

            for s in submissions:
                s.print_results()
            print() # Space between students


    def output_txt(self, file="output.txt"):
        """
        Creates .txt file output of test results
        
        Parameters:
            file (str): name/path of the output file
        """
        asgmt_dict = self.to_dict()

        try:
            with open(file, 'w', encoding="utf-8") as out:
                out.write(self.file_header() + "\n")

                for submissions in asgmt_dict.values():
                    points_earned = 0
                    for s in submissions:
                        points_earned += s.grade

                    problem_str = f"{submissions[0].student_name}:"
                    if submissions[0].is_canvas: # Only print if the student is known
                        problem_str += f" {points_earned:0.2f}/{self.max_points:0.2f}"
                        problem_str += f" ({100*points_earned/self.max_points:2.0f}%)\n"

                    out.write(problem_str)
                    for s in submissions:
                        s.print_results(out)
                    out.write("\n\n")
        except PermissionError as e:
            e.add_note(f"Output text document {file} is open or inaccessible")
            raise e


    def output_csv(self, file):
        """
        Creates .csv file output of test results
        
        Parameters:
            file (str): name/path of the output file
        """
        df = self.output_dataframe()

        try:
            with open(file, '+w', encoding="utf-8") as csv:
                csv.write(self.file_header())
                df.to_csv(csv, mode='a', lineterminator='\n', float_format='%.3f')
        except PermissionError as e:
            e.add_note(f"Output text document {file} is open or inaccessible")
            raise e


    def pypackage_command(self, target_file):
        """
        Develops a pyinstaller command based on required inputs

        Parameters:
            target_file (str): Python script being converted into a .exe
        """

        def update_path_list(all_paths, new_file):
            """
            Makes the file string, but also adds a path if needed
            """
            new_path = os.path.dirname(new_file)
            if new_path not in all_paths:
                all_paths.append(new_path)
            return all_paths


        path_list = []

        out_str = 'pyinstaller --noconfirm --onefile --console --contents-directory "."'
        out_str += f' --add-data "{self.config_file};."'

        for p in self.problems:
            out_str += f' --hidden-import "{utils.file_strip(p.solution_file)}"'
            path_list = update_path_list(path_list, p.solution_file)

            if p.runner_file:
                if p.runner_hidden:
                    out_str += f' --hidden-import "{utils.file_strip(p.runner_file)}"'
                    path_list = update_path_list(path_list, p.runner_file)
                else:
                    out_str += f' --add-data "{p.runner_file};."'

        if ico_dir := exist_icon():
            out_str += f' --icon "{ico_dir}"'

        if len(path_list) >= 0:
            out_str += f' --paths {":".join(path_list)}'

        out_str += f' "{target_file}"'
        return out_str


def exist_icon():
    """
    Checks if icon file can be found
    """
    return utils.in_repo('pygrader.ico')



"""
utils.py
weavebj1 - 2025

Basic repository needs

Part of python-grader library
"""

import os
import sys
import builtins
import logging

# These need to be separate imports because it works...
import matplotlib
import matplotlib.pyplot as plt
from matplotlib.axes import Axes
from matplotlib.figure import Figure


def repo_path():
    """
    Gets the path of the repository (breaks if this file moves)
    """
    return os.path.dirname(__file__)


def in_repo(search_file):
    """
    Returns the path of a file if it exists
    """
    search_dir = repo_path()
    for root, _, files in os.walk(search_dir):
        for file in files:
            if search_file == file:
                return os.path.join(root, file)
    return None


def running_in_exe():
    """
    Returns true if the code is running in an exe file
    """
    return getattr(sys, 'frozen', False) and hasattr(sys, '_MEIPASS')


def exe_resource_path():
    """
    Retrieves the path for files in the exe that are placed into a temp directory
    """
    return getattr(sys, '_MEIPASS')


def guarantee_path(path):
    """
    Makes sure that the file is on the system path
    """
    sys.path.append(os.path.dirname(os.path.abspath(path)))


def file_strip(path):
    """
    Removes all except for the base name (including a ./ at the start)
    """
    filename = os.path.basename(path)
    return os.path.splitext(filename)[0].replace('.','')


def verify_number_of_outputs(outputs, n):
    """
    Returns errors (and the correct number of elements)
    if the number of outputs is not the expected number
    
    Outputs: A tuple from some file
    n: The number of expected values in an output
    """

    if outputs is None or len(outputs) != n:
        e = ValueError("Returned improper number of outputs")
        return (e,)*4

    return outputs

def disable_breakpoints():
    """
    Disables any user breakpoints from running
    """
    sys.breakpointhook = nothing_function()


def log_verbose(message, *args, **kwargs):
    """
    Runs a log at an intermediate level
    """
    logging.log(logging.VERBOSE, message, *args, **kwargs)

# Diagnostic support
logging.addLevelName(logging.INFO-5, "VERBOSE")
logging.VERBOSE = logging.INFO-5
setattr(logging, 'verbose', log_verbose)

def disable_printing():
    """
    Disables just the print function
    """

    def new_print(*args, **kwargs):
        if len(args) > 0:
            log_verbose(*args, **kwargs)
        else:
            log_verbose("", *args, **kwargs)

    builtins.print = new_print

def close_plots():
    """
    Run once in a while to clear out student plots. Hacky solution
    """
    plt.close("all")


def disable_plotting():
    """
    Disables the plotting, figure display, and figure saving for matplotlib
    """

    matplotlib.use("agg")
    plt.close('all')
    Axes.plot = disable_function()
    Axes.legend = disable_function()
    Figure.savefig = disable_function()
    plt.show = disable_function()


def nothing_function(noutputs=0):
    """
    Does nothing. Used to replace breakpoint
    """
    return ([],)*noutputs


def disable_function(noutputs=0):
    """
    Replaces function with a dummy that does nothing

    noutputs: determines how many outputs the function will provide (empty)
    """

    def nope(*args, **kwargs):
        return nothing_function(noutputs=noutputs)

    return nope



"""
test_file.py
weavebj1 - 2025

File objects that run python files and store attributes

Part of python-grader library

TODO Add "import" type of file which can run multiple functions
"""

import os
import importlib
import re
import logging

from unittest.mock import Mock
import numpy as np
from mupygrader.utils import guarantee_path, file_strip


class TestFile:
    """
    TestFile object represents a file that is evaluated under testing conditions

    Attributes:
        filename (str): The name of the file under test
        filepath (str): The directory location of the file
        params (json_interface.problem): Problem definition derived from json file
        code (TestScript or TestFunction): [has an] object that represents the 
                                             specific code in the file 
    """

    def __init__(self, filepath, problem):
        """
        Constructor for TestFile object

        Parameters:
            filepath (str): Fully defined path of the file being used
            problem (json_interface.problem): Problem that this file is associated with 
        """
        split_path = filepath.split(os.path.sep)
        self.filename = split_path[-1]
        self.filepath = filepath
        self.params = problem

        try:
            if problem.type.lower() not in ["function", "script"]:
                raise ValueError
        except ValueError as e:
            e.add_note('Parameter "pytype" must be "function" or "script"')
            raise e
        if problem.type.lower() == "function":
            self.code = TestFunction(self)
        elif problem.type.lower() == "script":
            self.code = TestScript(self)

    def init_function_trackers(self):
        """
        Begins the tracking of any functions that are tracked by function tests
        """
        fun_tests = self.params.function_tests

        for fun in fun_tests:

            fun.mock = Mock()
            fun.mock.side_effect = fun.original

            setattr(fun.module, fun.function, fun.mock)

    def close_function_trackers(self):
        """
        Ends the tracking of any functions that are tracked by function tests
        """
        fun_tests = self.params.function_tests
        for fun in fun_tests:
            setattr(fun.module, fun.function, fun.original)

    def run_with_mock(self):
        """
        Runs the code of a file with tracking enabled
        """
        blank_array_size = np.size(self.params.function_tests)
        start_counts   = np.zeros(blank_array_size)
        end_counts     = np.zeros(blank_array_size)
        should_be_used = np.zeros(blank_array_size)

        # Get starting use-counts for functions
        idx = 0
        for fun in self.params.function_tests:
            start_counts[idx] = fun.mock.call_count
            should_be_used[idx] = not fun.invert
            idx += 1

        try:
            self.code.run()

        except Exception as e:
            e.add_note(f"Code failed while running {self.filename}")
            self.code.result = e
            return np.zeros(blank_array_size).tolist()

        finally:

            # Get final use counts for functions (up until any errors)
            idx = 0
            for fun in self.params.function_tests:
                end_counts[idx] = fun.mock.call_count
                idx += 1

        was_used = np.dot(1, end_counts > start_counts)
        function_scorecard = 1*(should_be_used == was_used)

        return function_scorecard.tolist()


class TestScript:
    """
    Represents the code for a file under test that is a script file

    Attributes:
        parent (TestFile): The name of the file under test
        result (tuple): Collected namespace from running the script
    """

    def __init__(self, file):
        """
        Constructor for TestScript

        Parameters:
            file (TestFile): Parent object for script file obj
        """
        self.parent = file
        self.result = []


    def run(self):
        """
        Runs TestScript file and saves result

        In a separate namespace:
            - if the module has already been imported, remove the key from the module
              list, forcing it to be reloaded when imported
            - import and store the result of the import to a variable, which includes
              any variables that existed in the script
        """
        sol_local_space = {"importlib":importlib}
        try:
            filename = file_strip(self.parent.filepath)
            guarantee_path(self.parent.filepath)
            exec('import sys', {}, sol_local_space) # pylint: disable=exec-used
            exec(f'if "{filename}" in sys.modules: del sys.modules["{filename}"]', # pylint: disable=exec-used
                 {},sol_local_space)
            exec(f'output = importlib.import_module("{filename}")',{},sol_local_space)  # pylint: disable=exec-used
            self.result = sol_local_space["output"]
        except Exception as e:
            e.add_note(f"Could not run script of file {filename}")
            self.result = e


    def compare_values(self, key):
        """
        Compares results of this object to another TestScript

        Parameters:
            key (TestScript): code object for solution code
        """
        scorecard = np.zeros(len(self.parent.params.value_tests)).tolist()
        i=0
        for var_test in self.parent.params.value_tests:
            var_name = str(var_test.variable)

            # Conditional to see if the value exists in general
            try:
                scorecard[i] = var_test.equal_tol(getattr(self.result, var_name),
                                                  getattr(key.result, var_name))
            except Exception as e:
                e.add_note(f"Failed while comparing values for {var_test.var_name}")
                scorecard[i] = e

            i+=1

        return scorecard



class TestFunction:
    """
    Represents the code for a file under test that is a function

    Attributes:
        parent (TestFile): The file under test
        result (tuple): Collected output returned by the function
    """

    def __init__(self, file):
        """
        Constructor for TestFunction

        Parameters:
            file (TestFile): Parent object for function obj
        """
        self.parent = file
        self.result = []


    def run(self):
        """
        Runs TestFunction file and saves result
        """
        fun_name = self.parent.params.function_name
        module_name = file_strip(self.parent.filename)
        guarantee_path(self.parent.filename)

        try:
            test_file = importlib.import_module(module_name)
            importlib.reload(test_file)
        except Exception as e:
            e.add_note(f"Failed import for {self.parent.filename}")
            self.result = e
            return

        if fun_name not in dir(test_file):
            raise ValueError(f'Function {fun_name} not found in {test_file.__name__}')
        fun_to_run = getattr(test_file, self.parent.params.function_name)

        if self.parent.params.runner:
            filename = file_strip(self.parent.params.runner.filename)
            guarantee_path(self.parent.params.runner.filename)
            runner_file = importlib.import_module(filename)
            runner_to_run = getattr(runner_file, 'main')
            self.result = runner_to_run(fun_to_run) # Pass the function obj, not the result
        else:
            self.result = fun_to_run()

    def compare_values(self, key):
        """
        Compares results of this object to another TestFunction

        Parameters:
            key (TestFunction): code object for solution code
        """

        scorecard = np.zeros(len(key.result)).tolist()
        i = 0
        for var_test in self.parent.params.value_tests:
            try:
                scorecard[i] = var_test.equal_tol(self.result[i], key.result[i])
            except Exception as e:
                e.add_note(f"Failed while comparing values for {var_test.variable}")
                scorecard[i] = e

            i+=1

        return scorecard



class Submission(TestFile):
    """
    Represents a file that is submitted by a student and will be compared
    against a solution

    Attributes:
        super(): TestFile class parameters
        student_name (str): Name of student
        student_id (str): Id number that canvas assigns to student
        is_late (bool): Based on canvas submission format. Affects grading
        is_canvas (bool): Flag based on detecting canvas format in docstring

        function_scorecard ([num]): Array showing performance on function tests
        value_scorecard ([num]): Array showing performance on value tests 
        scorecard ([num]): Composite of function and value scorecards

        points_scored (num): Sum of points gained, ignoring late penalties
        grade (num): Grade based on points and late penalties
        error_comments (str): List of feedback statements based on errors
    """

    def __init__(self, filepath, problem):
        """
        Constructor for Submission

        Parameters:
            filepath (str): File path to the file being submitted
            problem (json_interface.problem): Problem obj that submission is 
                                              associated with
        """
        super().__init__(filepath, problem)

        self._is_canvas = re.search(r"\S*_[^LATE_$]?(\d*)_\d+", self.filename)
        self.student_name = self.filename.split("_")[0]
        self.student_id = []
        self.is_late = False

        if self._is_canvas:
            self.student_id =  self._is_canvas.group(1)
            self.is_late = self.filename.split("_")[1].__contains__("LATE")
            self.is_canvas = True
        else:
            self.is_canvas = False

        self.function_scorecard = []
        self.value_scorecard = []
        self.scorecard = []
        self.points_scored = 0
        self.grade = 0
        self.error_comments = ""

    def update_score(self):
        """
        Updates score and grade based on results from running code
        """
        self.points_scored = 0
        i = 0
        for t in self.params.all_tests:
            if not isinstance(self.scorecard[i], Exception) and (self.scorecard[i] > 0):
                self.points_scored += t.points
            i+=1


        multiplier = 1
        if self.is_late:
            multiplier *= self.params.assignment.late_mult

        self.grade = self.points_scored * multiplier

    def make_error_msg(self):
        """
        Creates error messages based on tests failed
        """
        messages = []
        if isinstance(self.code.result, Exception):
            error_type = type(self.code.result).__name__
            messages.append((
                f" {error_type}: {self.code.result} \n"
                f" {"\n ".join(self.code.result.__notes__)}"))

        i = 0
        for t in self.params.all_tests:
            if isinstance(self.scorecard[i], Exception):
                messages.append((
                    f"[-{t.points:.3f}] ERROR {self.scorecard[i]} \n"
                    f" {"\n".join(self.scorecard[i].__notes__)}"))
            elif self.scorecard[i] == 0:
                messages.append(f"[-{t.points:.3f}] {t.error_comment}")
            elif np.isnan(self.scorecard[i]):
                messages.append(f"[-{t.points:.3f}] A test ({t.name}) could not be evaluated")
            i+=1

        self.error_comments = "\n".join(messages) + "\n"

    def run_and_score(self, solution):
        """
        Evaluates submission based on a given solution file

        Parameters:
            solution (TestFile): Solution being compared against
        """
        self.init_function_trackers()
        self.function_scorecard = self.run_with_mock()
        self.close_function_trackers()

        if not isinstance(self.code.result, Exception):
            try:
                self.value_scorecard = self.code.compare_values(solution.code)
            except Exception as e:
                self.value_scorecard = [0]*len(self.params.value_tests)
                raise ValueError(f"Could not compare any values for {self.filename}") from e
        else:
            self.value_scorecard = [0]*len(self.params.value_tests)

        self.scorecard = self.value_scorecard + self.function_scorecard
        self.update_score()
        self.make_error_msg()


    def print_results(self, fileobj=''):
        """
        Prints the results to python terminal or to a file if given

        Parameters:
            fileobj (str): File that results are written to (optional)
        """
        strs = []
        strs.append((f"Grader gave {self.grade:05.2f}/"
                    f"{self.params.points:05.2f} for {self.filename}"))
        if self.is_late:
            strs.append(f"[-{self.points_scored-self.grade:.3f}] File was submitted late")
        strs.append(self.error_comments)

        outstr = "\n".join(strs)

        if fileobj:
            fileobj.write(outstr + "\n")
        else:
            logging.info(outstr)


    @property
    def identifier(self):
        """
        Returns a unique identifier for output dictionaries
        """
        if self.is_canvas:
            return (self.student_name, self.student_id)
        else:
            return self.student_name


def get_pyfiles(folder):
    """
    Helper function to get files but ignore __pycache__ directories
    that are created due to importing of modules
    """

    pyfiles = []
    test_files = os.listdir(folder)

    if '__pycache__' in test_files:
        test_files.remove('__pycache__') # Directory added due to import statements

    for f in test_files:
        if f.endswith(".py"):
            pyfiles.append(f)

    return pyfiles



"""
json_interface.py
weavebj1 - 2025

Reads json files that contain testing parameters

Part of python-grader library
"""

import importlib
import numpy as np
from mupygrader import test_file


class TestInfo:
    """
    TestInfo object represents a test with conditions

    Attributes:
        name (str): The name of the test
        error_comment (str): A statement given when the answer is incorrect
        weight (num): Relative weight of this test relative to others within a problem
        points (num): Calculated value of test based on problem's total points and weight
        invert (bool): By default, checks that test passed. If inverted, makes sure that it doesn't 
    """

    def __init__(self, name, error_comment, weight, invert):
        """
        Constructor for test object

        Parameters:
            [See attributes above. Points is a derived attribute]
        """
        self.name = name
        self.error_comment = error_comment
        self.weight = weight
        self.points = []
        self.invert = invert


class ValueTest(TestInfo):
    """
    Constructor for ValueTest object

    Attributes:
        abs_tol (num): Linear slop in results that is allowable (+/- diff)
        rel_tol (num): Multiplicative slop in results that is allowable (percent diff)
        [See TestInfo for parent attributes]
    """

    def __init__(self, variable, abs_tol=0, rel_tol=0, error_comment="",
                 test_weight=1, invert=False):
        """
        Constructor for ValueTest object

        Parameters:
            [See Attributes for ValueTest]
        """
        if len(error_comment) == 0:
            error_comment = f"Variable '{variable}' had incorrect value"

        super().__init__(variable, error_comment, test_weight, invert)
        self.variable = variable
        self.abs_tol = max(abs_tol, 1e-9) # Default for math.isclose() for floating point errors
        self.rel_tol = rel_tol


    def equal_tol(self, submission_val, key_val):
        """
        Compares two values based on tolerances in a value test
        If both tolerances are given, math.isclose is generous 
            (either can be true, both are not required to pass)

        Parameters:
            submission_val (any?): Value from the submission
            key_val (any?): Value from the answer key
        """

        # Check length of any tuples first
        if isinstance(key_val, tuple) and len(key_val) != len(submission_val):
            raise ValueError((f"For {self.variable}, key and submission had different "
                              f"number of outputs ({len(key_val)}, {len(submission_val)})"))
        
        # Do this whole function recursively if the value is a tuple
        if isinstance(key_val, tuple):
            tuple_matches = True
            for sub,key in zip(submission_val,key_val):
                tuple_matches &= self.equal_tol(sub, key)
            return tuple_matches

        if isinstance(submission_val, Exception):
            return False

        # Check shapes first so that everything below can assume the same length
        if np.shape(key_val) != np.shape(submission_val):
            raise ValueError((f"For {self.variable}, key and submission had different "
                              f"shapes ({np.shape(key_val)}, {np.shape(submission_val)})"))

        check_format = "unknown"

        # Check to see if the expected value is a numpy datatype
        if isinstance(key_val, np.ndarray) or isinstance(key_val, np.generic):
            if key_val.dtype.kind in ['b','f','i']:
                check_format = "numeric"
            elif np.issubdtype(key_val.dtype, np.character):
                check_format = "string"
        else:
            if isinstance(key_val, (int, float, bool)):
                check_format = "numeric"
            elif isinstance(key_val, (str, list)):
                check_format = "string"
            elif key_val is None:
                check_format = "none"


        match check_format:
            case "numeric":
                result = np.allclose(submission_val, key_val,
                                     rtol=self.rel_tol,
                                     atol=self.abs_tol,
                                     equal_nan=True)
            case "string":
                result = np.all(submission_val == key_val)
            case "none":
                result = submission_val == key_val
            case "unknown":
                raise ValueError("Tests for this type of data have not been set up")

        return result


class FunctionTest(TestInfo):
    """
    Constructor for FunctionTest object

    Attributes:
        function (str): Name of function from module
        module (str): Name of module that function is derived from
        [See TestInfo for parent attributes]
    """
    def __init__(self, function, module, error_comment="", test_weight=1, invert=False):
        """
        Constructor for FunctionTest object

        Parameters:
            [See attributes for FunctionTest]
        """

        if len(error_comment) == 0 and not invert:
            error_comment = f"Function '{function}' from '{module}' was not used"
        elif len(error_comment) == 0 and invert:
            error_comment = f"Function '{function}' from '{module}' was used, but should not be"

        super().__init__(function, error_comment, test_weight, invert)

        self.module = module
        self.function = function
        self.original = []
        self.mock = []
        try:
            self.module = importlib.import_module(self.module, function)
            self.original = getattr(self.module, function)
        except ModuleNotFoundError as e:
            e.add_note((f'Could not import from "{self.module}". Check that it is installed and '
                   'that its full name is spelled out correctly'))
            raise e
        except AttributeError as e:
            e.add_note((f'"{function}" from "{self.module}" could not import. Check the '
                   'name of the function and make sure that all items are installed'))
            raise e


class Problem:
    """
    Problem Object defines problem information from json files

    Attributes:
        assignment (assignment.Assignment): Parent assignment
        name (str): Name of problem for bookkeeping
        file_whitelist (str): Filename search filter for this problem
        solution_file (str): Name of file used as solution
        solution (test_file.TestFile): obj for solution file
        runner_file (str): Name of file used to run FunctionTest
        runner_hidden (bool): Keeps track of the runner file being available in .exe
        runner (test_file.TestFile): obj for runner file
        type (str): Either "script" or "function"
        points (num): Value of problem in points
        value_tests ([test_file.ValueTest]): List of value tests to assess 
        function_tests ([test_file.FunctionTest]): List of function tests to assess
        submissions ([test_file.Submission]): List of student submissions associated with problem
    """

    def __init__(self, asgmt, name, file_whitelist, solution_file, function_name='', runner_file='',
                 runner_hidden=False, pytype="script", prob_points=0, returned_vars=None,
                 used_functions=None):
        """
        Constructor for Problem object

        Parameters:
            asgmt (assignment.Assignment): Parent assignment
            name (str): Name of problem for bookkeeping
            file_whitelist (str): Filename search filter for this problem
            solution_file (str): Name of file used as solution
            function_name (str): Name of function to run from file
            runner_file (str): Name of file used to run FunctionTest
            runner_hidden (bool): Keeps track of the runner file being available in .exe
            pytype (str): Either "script" or "function"
            prob_points (num): Value of problem in points
            returned_vars ([dict]): List of value tests to assess (from json) 
            used_function ([dict]): List of function tests to assess (from json)
        """
        self.assignment = asgmt
        self.name = name
        self.file_whitelist = file_whitelist
        self.solution_file = solution_file
        self.function_name = function_name
        self.solution = []
        self.runner_file = runner_file
        self.runner_hidden = runner_hidden
        self.runner = []
        self.type = pytype
        self.points = prob_points

        self.value_tests = []
        if returned_vars:
            for t in returned_vars:
                self.value_tests.append(ValueTest(**t))

        self.function_tests = []
        if used_functions:
            for t in used_functions:
                self.function_tests.append(FunctionTest(**t))

        total_weights = sum(self.function_weights + self.value_weights)
        for t in self.all_tests:
            t.points = prob_points * t.weight / total_weights

        self.submissions = []


    def init_solution(self):
        """
        Creates the solution TestFile object based on a found solution file
        """
        self.solution = test_file.TestFile(self.solution_file, self)


    def init_runner(self):
        """
        Creates the runner TestFile object based on a found runner file
        """
        self.runner = test_file.TestFile(self.runner_file, self)


    def add_submission(self, file):
        """
        Creates a submission TestFile object and adds it to the list
        """
        self.submissions.append(test_file.Submission(file, self))

    @property
    def all_tests(self):
        """
        Returns all of the value and function tests in the problem
        """
        return [*self.value_tests, *self.function_tests]

    @property
    def value_weights(self):
        """
        Returns the weights of all value tests
        """
        weights = []
        for w in self.value_tests:
            weights.append(w.weight)
        return weights

    @property
    def function_weights(self):
        """
        Returns the weights of all function tests
        """
        weights = []
        for w in self.function_tests:
            weights.append(w.weight)
        return weights



import tkinter
from tkinter import ttk
import eel
import sv_ttk
import darkdetect

root = tkinter.Tk()

button = ttk.Button(root, text="Click me!")
button.pack()



sv_ttk.set_theme(darkdetect.theme())

root.mainloop()